PutricidePlague = LibStub("AceAddon-3.0"):NewAddon("PutricidePlague", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0");
local core = PutricidePlague;
local L = LibStub("AceLocale-3.0"):GetLocale("PutricidePlague");

local defaults = {
	profile = {
		locked = false,
		pos = {x=500, y=300},
		arrows = true,
		whispers = false,
	}
};

local GREEN = {r=0, g=1, b=0};
local RED = {r=1, g=0, b=0};
local WHITE = {r=1, g=1, b=1};
local YELLOW = {r=1, g=1, b=0};
local BROWN = {r=0.5, g=0.25, b=0};
local PLAGUE = GetSpellInfo(70911);
local SICKNESS = GetSpellInfo(73117);
local MODE_SETTINGS = {
	[3] = { -- 10 player heroic
		passTime = 12,
		safeThresh = 15,
	},
	[4] = { -- 25 player heroic
		passTime = 8,
		safeThresh = 12,
	},
};
local PLAYER = {
	RANGED = 1,
	MELEE = 2,
	TANK = 3,
};
local ROOM = {w = 1069.6156745738, h = 713.83371679543}; -- Stolen from DXE (Coords.lua)
local addon, Arrows;

core.playerStart = {};
core.first = false;
core.dispRegistry = {};
core.openTexts = 0;
core.numTexts = 0;

--[[FOR DEBUG ONLY
local PLAGUE = "Lifebloom";
local SICKNESS = "Lifebloom-";
local function UnitDebuff(...)
	return UnitAura(...);
end--]]

function core:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("PutricidePlagueDB", defaults, "Default");
	self:RegisterChatCommand("pp", "SlashProcess");
	self:RegisterChatCommand("putriplague", "SlashProcess");
	self:RegisterChatCommand("putricideplague", "SlashProcess");
	
	self:CreateAnchor();
end

function core:OnEnable()
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "ZoneChange");
	self:RegisterEvent("ZONE_CHANGED", "ZoneChange");
	self:RegisterEvent("PLAYER_ENTERING_WORLD", "ZoneChange");
	
	self:LoadPos();
	if not self.db.profile.locked then
		self.anchorFrame:SetBackdropColor(0.1, 0.1, 0.1, 0.75);
		self.anchorFrame:EnableMouse(true);
	end	
	addon = DXE;
	if (addon) then
		Arrows = addon.Arrows;
	end
end

function core:OnDisable()

end

function core:SlashProcess(msg)
	if msg == "lock" then
		if self.db.profile.locked == true then
			self.anchorFrame:SetBackdropColor(0.1, 0.1, 0.1, 0.75);
			self.anchorFrame:EnableMouse(true);
			self.db.profile.locked = false;
			self:Print(L["Anchor frame unlocked"]);
		else
			self.anchorFrame:SetBackdropColor(0.1, 0.1, 0.1, 0);
			self.anchorFrame:EnableMouse(false);
			self.db.profile.locked = true;
			self:Print(L["Anchor frame locked."]);
		end
	elseif msg == "reset" then
		self.db.profile.pos = {x=500, y=300};
		self:LoadPos();
		self.anchorFrame:SetBackdropColor(0.1, 0.1, 0.1, 0.75);
		self.anchorFrame:EnableMouse(true);
		self.db.profile.locked = false;
		self:Print(L["Anchor frame reset and unlocked."]);
	elseif msg == "clear" then
		self:ZoneChange();
		self:Print(L["Data and display cleared."]);
	elseif msg == "arrows" then
		if self.db.profile.arrows then
			self.db.profile.arrows = false;
			self:Print(L["DXE Arrows disabled."]);
		else
			self.db.profile.arrows = true;
			self:Print(L["DXE Arrows enabled."]);
		end
	elseif msg == "whispers" then
		if self.db.profile.whispers then
			self.db.profile.whispers = false;
			self:Print(L["Target whispering disabled."]);
		else
			self.db.profile.whispers = true;
			self:Print(L["Target whispering enabled."]);
		end
	else
		self:Print(L["Valid commands: lock, reset, clear, arrows, whispers"]);
	end
end

function core:CreateAnchor()
	if not self.anchorFrame then
		self.anchorFrame = CreateFrame("Frame", nil, UIParent);
		self.anchorFrame:SetWidth(150);
		self.anchorFrame:SetHeight(20);
		self.anchorFrame:SetBackdrop({
			bgFile = "Interface\\Tooltips\\UI-Tooltip-Background", tile = true, tileSize = 8,
		});
		self.anchorFrame:SetBackdropColor(0.1, 0.1, 0.1, 0.75);
		self.anchorFrame:SetMovable(true);
		self.anchorFrame:RegisterForDrag("LeftButton");
		self.anchorFrame:SetScript("OnDragStart", function() if not self.db.profile.locked then self.anchorFrame:StartMoving() end end);
		self.anchorFrame:SetScript("OnDragStop", function() self.anchorFrame:StopMovingOrSizing() self:SavePos() end);
		if self.db.profile.locked == true then
			self.anchorFrame:SetBackdropColor(0.1, 0.1, 0.1, 0);
			self.anchorFrame:EnableMouse(false);
		else
			self.anchorFrame:SetBackdropColor(0.1, 0.1, 0.1, 0.75);
			self.anchorFrame:EnableMouse(true);
		end
	end
end

function core:LoadPos()
	local x = self.db.profile.pos.x;
	local y = self.db.profile.pos.y;
	local s = self.anchorFrame:GetEffectiveScale();
	self.anchorFrame:ClearAllPoints();
	self.anchorFrame:SetParent(UIParent);
	self.anchorFrame:ClearAllPoints();
	self.anchorFrame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", x / s, y / s);
end

function core:SavePos()
	local s = self.anchorFrame:GetEffectiveScale();
	self.db.profile.pos.x = self.anchorFrame:GetLeft()*s;
	self.db.profile.pos.y = self.anchorFrame:GetTop()*s;
end

function core:COMBAT_LOG_EVENT_UNFILTERED(_, timestamp, event, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellid, spellName, spellSchool, amount, ...)
	if (spellName == PLAGUE) then
		if (event == "SPELL_AURA_APPLIED" and UnitExists(dstName)) then
			self.playerStart[dstName] = GetTime();
		end
		if (event == "SPELL_AURA_REMOVED") then
			self.playerStart[dstName] = nil;
		end
		self:UpdateTimers(not event:find("_AURA"));
	end
	if (spellName == SICKNESS or event == "UNIT_DIED") then
		self:UpdatePriority();
	end
end

function core:ZoneChange()
	self.first = false;
	self.playerStart = {};
	self:ClearDisplay();
	
	if (GetRealZoneText() == L["Icecrown Citadel"]) then
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	else
		self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
	end
end

function core:UpdateTimers(tick)
	if (not self.first) then
		self:InitTimers();
	end
	local found = false;
	local color = WHITE;
	local mode = self:GetMode();
	local opts = MODE_SETTINGS[mode];
	if (not opts) then error("Bad instance mode."); return; end
	
	for i=1,GetNumRaidMembers() do
		local name, _, _, _, _, duration, expires = UnitDebuff("raid"..i, PLAGUE);
		if (name) then 
			local sick, _, _, stack = UnitDebuff("raid"..i, SICKNESS);
			local warn = "";
			if (sick) then
				for j=1, stack do
					warn = warn .. "!";
				end
				color = RED;
			end
			local pName = UnitName("raid"..i);
			local pTime, tTime = floor(GetTime()-(self.playerStart[pName] or 0)+0.5), floor(expires-GetTime()+0.5);
			local time = format("%s%s %ds / %ds", UnitName("raid"..i), warn, pTime, tTime);
			if ((pTime+tTime)*(1+2.5*(stack or 0)) <= opts.safeThresh) then
				color = GREEN;
			else
				if (tick and IsRaidOfficer() and self.db.profile.whispers) then
					if (pTime >= opts.passTime-2 and pTime<=opts.passTime) then
						SendChatMessage(format(">>> "..L["Pass to {star} in %d"].." <<<", opts.passTime-pTime), "whisper", nil, UnitName("raid"..i));
						SendChatMessage(format(L["Pass to {star} in %d"], opts.passTime-pTime), "raid");
					end
					if (pTime >= opts.passTime+2 or color == RED) then
						SendChatMessage(L[">>> PASS TO {star} NOW <<<"], "whisper", nil, UnitName("raid"..i));
						SendChatMessage(L["PASS TO {star} NOW"], "raid");
					end
				end
			end
			self:UpdateText(1, time, color);
			-- self:Print(time, self.playerStart[UnitName("raid"..i)], UnitName("raid"..i), expires, GetTime());
			found = true;
			break;
		end
	end
	if (not found) then
		self:UpdateText(1, "", WHITE);
	end
	self:UpdatePriority();
	self:ScheduleTimer("UpdatePriority", 0.5)
end

function core:UpdatePriority()
	if (not self.first) then
		return;
	end
	local mode = self:GetMode();
	local opts = MODE_SETTINGS[mode];
	if (not opts) then return; end
	
	local target = next(self.playerStart);
	if(not target) then
		-- clear and quit
		for i=2,4 do
			self:UpdateText(i, "");
		end
		for i=1,GetNumRaidMembers() do
			if(GetRaidTargetIndex("raid"..i) == 1) then
				SetRaidTarget("raid"..i, 0);
			end
		end
		return;
	end
	
	-- double check
	if (not UnitDebuff(target, PLAGUE)) then
		-- unset and restart, something bugged.
		self.playerStart[target] = nil;
		self:UpdatePriority();
		return;
	end
	
	-- Find current coords
	local px, py;
	SetMapToCurrentZone();
	for i=1,GetNumRaidMembers() do
		if (UnitName("raid"..i) == target) then
			px,py = GetPlayerMapPosition("raid"..i);
		end
	end
	-- self:Print("Self at: ", px, py);
	
	local prox = {};
	for i=1,GetNumRaidMembers() do
		local name, _, _, _, _, _, zone = GetRaidRosterInfo(i);
		--print((select(7,UnitDebuff("raid"..i, SICKNESS)) or 0), "<", self.playerStart[target]+opts.safeThresh);
		if  ( select(7,UnitDebuff("raid"..i, SICKNESS)) and 
			((select(7,UnitDebuff("raid"..i, SICKNESS)) or 0)<self.playerStart[target]+opts.safeThresh) ) then
			--self:Print("Plague on but will be free: ", name);
		end
		if (
			zone == L["Icecrown Citadel"] and
			((select(7,UnitDebuff("raid"..i, SICKNESS)) or 0)<self.playerStart[target]+opts.safeThresh)  and -- if sickness (or lack thereof) will expire before a safe pass (this also pads it by a few seconds)
			not UnitDebuff("raid"..i, PLAGUE) and 
			not UnitDebuff("raid"..i, GetSpellInfo(72858)) and
			not UnitDebuff("raid"..i, GetSpellInfo(72838)) and
			not UnitIsDeadOrGhost("raid"..i) and
			UnitIsConnected("raid"..i) and
			not UnitInVehicle("raid"..i)
			) then
			local dx,dy = GetPlayerMapPosition("raid"..i);
			local diff = math.sqrt(((px-dx)*ROOM.w)^2+((py-dy)*ROOM.h)^2);
			-- self:Print("Adding:", name, dx, dy, diff);
			table.insert(prox, {["name"]=name, ["diff"]=diff, ["type"]=self:GetPlayerType(name)});
		end
	end
	table.sort(prox, function(a,b)
		if (a.type ~= b.type) then
			return a.type < b.type;
		end
		return a.diff < b.diff;
	end);
	for i,v in ipairs(prox) do
		-- self:Print("", "", v.name, v.diff);
	end
	for i=2,4 do
		if(prox[i-1]) then
			local color = (prox[i-1].type==PLAYER.MELEE and YELLOW) or (prox[i-1].type==PLAYER.TANK and BROWN) or WHITE;
			self:UpdateText(i, format(L["  %s (%d yds)"], prox[i-1].name, prox[i-1].diff), color);
			if(i == 2) then
				-- self:Print("SetRaidTarget", prox[i-1].name, 1);
				if(IsRaidOfficer() and GetRaidTargetIndex(prox[i-1].name) ~= 1) then
					SetRaidTarget(prox[i-1].name, 1);
				end
				if(self.db.profile.arrows) then
					if (not Arrows) then
						addon = DXE;
						if (addon) then
							Arrows = addon.Arrows;
						end
					end
					
					if (Arrows) then
						for p=1,GetNumRaidMembers() do
							Arrows:RemoveTarget(UnitName("raid"..p));
						end
						if(prox[i-1].name == UnitName("player")) then
							Arrows:RemoveTarget(target);
							Arrows:AddTarget(target, 10, "TOWARD", L["Get the plague from here"], "Position");
						end
						if(target == UnitName("player")) then
							Arrows:RemoveTarget(prox[i-1].name);
							Arrows:AddTarget(prox[i-1].name, 10, "TOWARD", L["Give the plague here"], "Position");
						end
					end
				end
			end
		else
			self:UpdateText(i, "NONE", RED);
		end
	end
end

function core:GetPlayerType(unit)
	local class, health, mana = select(2, UnitClass(unit)), UnitHealthMax(unit), UnitManaMax(unit);
	if (health > 50000) then
		return PLAYER.TANK;
	end
	if (class == "PALADIN" or class == "SHAMAN") then
		if (mana < 20000) then
			return PLAYER.MELEE;
		end
		return PLAYER.RANGED;
	elseif (class == "DRUID") then
		if (UnitHasMana(unit)) then
			return PLAYER.RANGED;
		else
			return PLAYER.MELEE;
		end
	elseif (class == "ROGUE" or class == "DEATHKNIGHT" or class == "WARRIOR") then
		return PLAYER.MELEE;
	end
	return PLAYER.RANGED;
end

function core:GetMode()
	local _, _, diff, _, _, dynHeroic, dynFlag = GetInstanceInfo(); 
	return (dynFlag and (2-(diff%2)+2*dynHeroic)) or diff;
end

function core:InitTimers()
	self:AddText(1, "")
	self:AddText(2, "")
	self:AddText(3, "")
	self:AddText(4, "")
	self.first = true;
end

function core:AddText(id, text, color)
	if self.openTexts < self.numTexts then 
		self.openTexts = self.openTexts + 1
		self.dispRegistry[self.openTexts]:SetText(text)
		self.dispRegistry[self.openTexts]:SetTextColor(1, 1, 1)
		self.dispRegistry[self.openTexts]:Show()
		self.dispRegistry[self.openTexts].regID = id
	else
		self.openTexts = self.openTexts + 1
		self.numTexts = self.numTexts + 1
		
		local obj = self.anchorFrame:CreateFontString(nil, "ARTWORK")
		obj:SetFontObject(TextStatusBarText)
		obj:SetJustifyH("LEFT")
		local labelFont, size, flags = GameFontHighlightMedium:GetFont()
		obj:SetFont(labelFont, size, "OUTLINE")
		obj:ClearAllPoints()
		obj:SetPoint("TOPLEFT", self.anchorFrame, "TOPLEFT", 0, -(self.numTexts-1)*16)
		if color == nil then
			obj:SetTextColor(1, 1, 1)
		else
			obj:SetTextColor(color.r, color.g, color.b)
		end
		obj:SetText(text)

		obj.regID = id
		table.insert(self.dispRegistry, obj)
	end
end

function core:UpdateText(id, val, color)
	if (not color) then
		color = WHITE;
	end
	for i,v in ipairs(self.dispRegistry) do
		if v.regID == id then
			v:SetText(val)
			v:SetTextColor(color.r, color.g, color.b)
			return
		end
	end
end

function core:ClearDisplay()
	for i,v in ipairs(self.dispRegistry) do
		v:Hide()
	end
	self.openTexts = 0
end

-- hide outgoing messages
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", function(self, event, msg, target)
	if (msg:find("^>>>.+<<<$")) then
		return true;
	end
end);
